#include <hermit.h>
#include <htypes.h>
#include <herrno.h>
#include <arch/system.h>
#include <arch/pinmux.h>
#include <rcm2_gpio.h>

#define INVALID_GPIO_BASE (0xffffffff)

static u32 get_gpio_base(int index)
{
	switch (index) {
	case 0:	return GPIO0_BASE_ADDR;
	case 1:	return GPIO1_BASE_ADDR;
	case 2:	return GPIO2_BASE_ADDR;
	case 3:	return GPIO3_BASE_ADDR;
	case 4:	return GPIO4_BASE_ADDR;
	case 5:	return GPIO5_BASE_ADDR;
	case 6:	return GPIO6_BASE_ADDR;
	case 7:	return GPIO7_BASE_ADDR;
	default: break;
	}
	return INVALID_GPIO_BASE;
}

int __maybe_unused gpio_direction_input(u32 pin)
{
	u32 index, mask, base, val;

	index = _PIN2GPINDEX(pin);
	mask = PIN2GPMASK(pin);

	base = get_gpio_base(index);
	if (base == INVALID_GPIO_BASE)
		return -H_EINVAL;

	val = read32(base + GPIO_INOUTSEL);
	val &= ~mask;
	write32(base + GPIO_INOUTSEL, val);

	return 0;
}

int __maybe_unused gpio_direction_output(u32 pin, u8 level)
{
	u32 index, mask, base, val;

	index = _PIN2GPINDEX(pin);
	mask = PIN2GPMASK(pin);

	base = get_gpio_base(index);
	if (base == INVALID_GPIO_BASE)
		return -H_EINVAL;

	val = read32(base + GPIO_INOUTSEL);
	val |= mask;
	write32(base + GPIO_INOUTSEL, val);

	val = read32(base + GPIO_OUTDT);
	if (level)
		val |= mask;
	else
		val &= ~mask;
	write32(base + GPIO_OUTDT, val);

	return 0;
}

int __maybe_unused gpio_get_value(u32 pin)
{
	u32 index, mask, base, val;

	index = _PIN2GPINDEX(pin);
	mask = PIN2GPMASK(pin);

	base = get_gpio_base(index);
	if (base == INVALID_GPIO_BASE)
		return -H_EINVAL;

	val = read32(base + GPIO_INDT);

	return !!(val & mask);
}

int __maybe_unused gpio_set_value(u32 pin, u8 level)
{
	u32 index, mask, base, val;

	index = _PIN2GPINDEX(pin);
	mask = PIN2GPMASK(pin);

	base = get_gpio_base(index);
	if (base == INVALID_GPIO_BASE)
		return -H_EINVAL;

	val = read32(base + GPIO_OUTDT);
	if (level)
		val |= mask;
	else
		val &= ~mask;
	write32(base + GPIO_OUTDT, val);

	return 0;
}
